/*
 *  Copyright (C) 1999-2014 Jens Thoms Toerring
 *
 *  This file is part of fsc2.
 *
 *  Fsc2 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3, or (at your option)
 *  any later version.
 *
 *  Fsc2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "rs690.h"


/*-----------------------------------------------------------------*
 * Converts a time into the internal type of a time specification,
 * i.e. an integer multiple of the time base
 *-----------------------------------------------------------------*/

Ticks
rs690_double2ticks( double p_time )
{
    double ticks;


    if ( ! rs690.is_timebase )
    {
        print( FATAL, "Can't set a time because no pulser time base has been "
               "set.\n" );
        THROW( EXCEPTION );
    }

    ticks = p_time / rs690.timebase;

    if ( ticks > TICKS_MAX || ticks < TICKS_MIN )
    {
        print( FATAL, "Specified time is too long for time base of %s.\n",
               rs690_ptime( rs690.timebase ) );
        THROW( EXCEPTION );
    }

    if (    fabs( Ticksrnd( ticks ) - p_time / rs690.timebase ) > 1.0e-2
         || ( p_time > 0.99e-9 && Ticksrnd( ticks ) == 0 ) )
    {
        print( FATAL, "Specified time of %s is not an integer multiple of the "
               "fixed pulser the time base of %s.\n",
               rs690_ptime( p_time ), rs690_ptime( rs690.timebase ) );
        THROW( EXCEPTION );
    }

    return Ticksrnd( ticks );
}


/*-----------------------------------------------------*
 * Does the exact opposite of the previous function...
 *-----------------------------------------------------*/

double
rs690_ticks2double( Ticks ticks )
{
    if ( ! rs690.is_timebase )
    {
        print( FATAL, "Can't set a time because no pulser time base has been "
               "set.\n" );
        THROW( EXCEPTION );
    }

    return rs690.timebase * ticks;
}


/*-----------------------------------------------*
 * Returns the structure for pulse numbered pnum
 *-----------------------------------------------*/

Pulse_T *
rs690_get_pulse( long pnum )
{
    Pulse_T *cp = rs690.pulses;


    if ( pnum < 0 )
    {
        print( FATAL, "Invalid pulse number: %ld.\n", pnum );
        THROW( EXCEPTION );
    }

    while ( cp != NULL )
    {
        if ( cp->num == pnum )
            break;
        cp = cp->next;
    }

    if ( cp == NULL )
    {
        print( FATAL, "Referenced pulse #%ld does not exist.\n", pnum );
        THROW( EXCEPTION );
    }

    return cp;
}


/*----------------------------------------------------*
 *----------------------------------------------------*/

const char *
rs690_ptime( double p_time )
{
    static char buffer[ 3 ][ 128 ];
    static size_t i = 2;


    i = ( i + 1 ) % 3;
    if ( fabs( p_time ) >= 1.0 )
        sprintf( buffer[ i ], "%g s", p_time );
    else if ( fabs( p_time ) >= 1.e-3 )
        sprintf( buffer[ i ], "%g ms", 1.e3 * p_time );
    else if ( fabs( p_time ) >= 1.e-6 )
        sprintf( buffer[ i ], "%g us", 1.e6 * p_time );
    else
        sprintf( buffer[ i ], "%g ns", 1.e9 * p_time );

    return buffer[ i ];
}


/*----------------------------------------------------*
 *----------------------------------------------------*/

const char *
rs690_pticks( Ticks ticks )
{
    return rs690_ptime( rs690_ticks2double( ticks ) );
}


/*-------------------------------------------------------------------*
 * Comparison function for two pulses: returns 0 if both pulses are
 * inactive, -1 if only the second pulse is inactive or starts at a
 * later time and 1 if only the first pulse is inactive pulse or the
 * second pulse starts earlier.
 *-------------------------------------------------------------------*/

int
rs690_pulse_compare( const void * A,
                     const void * B )
{
    Pulse_Params_T *a = ( Pulse_Params_T * ) A,
                   *b = ( Pulse_Params_T * ) B;

    return a->pos <= b->pos ? -1 : 1;
}


/*----------------------------------------------------*
 *----------------------------------------------------*/

void
rs690_show_pulses( void )
{
    int pd[ 2 ];
    pid_t pid;


    if ( pipe( pd ) == -1 )
    {
        if ( errno == EMFILE || errno == ENFILE )
            print( FATAL, "Failure, running out of system resources.\n" );
        return;
    }

    if ( ( pid =  fork( ) ) < 0 )
    {
        if ( errno == ENOMEM || errno == EAGAIN )
            print( FATAL, "Failure, running out of system resources.\n" );
        return;
    }

    /* Here's the childs code */

    if ( pid == 0 )
    {
        char * volatile cmd = NULL;

        close( pd[ 1 ] );

        if ( dup2( pd[ 0 ], STDIN_FILENO ) == -1 )
        {
            goto filter_failure;
            close( pd[ 0 ] );
        }

        close( pd[ 0 ] );

        TRY
        {
            cmd = get_string( "%s%sfsc2_pulses", bindir, slash( bindir ) );
            execl( cmd, "fsc2_pulses", NULL );
        }

    filter_failure:

        if ( cmd != NULL )
            T_free( cmd );
        _exit( EXIT_FAILURE );
    }

    /* And finally the code for the parent */

    close( pd[ 0 ] );
    rs690.show_file = fdopen( pd[ 1 ], "w" );
}


/*----------------------------------------------------*
 *----------------------------------------------------*/

void
rs690_dump_pulses( void )
{
    char *name;
    char *m;
    struct stat stat_buf;


    do
    {
        TRY
        {
            name = T_strdup( fsc2_show_fselector( "File for dumping pulses:",
                                                  NULL, "*.pls", NULL ) );
            TRY_SUCCESS;
        }
        OTHERWISE
            return;

        if ( name == NULL || *name == '\0' )
        {
            if ( name != NULL )
                T_free( name );
            return;
        }

        if  ( 0 == stat( name, &stat_buf ) )
        {
            m = get_string( "The selected file does already exist:\n%s\n"
                            "\nDo you really want to overwrite it?", name );
            if ( 1 != show_choices( m, 2, "Yes", "No", NULL, 2 ) )
            {
                T_free( m );
                name = T_free( name );
                continue;
            }
            T_free( m );
        }

        if ( ( rs690.dump_file = fopen( name, "w+" ) ) == NULL )
        {
            switch( errno )
            {
                case EMFILE :
                    show_message( "Sorry, you have too many open files!\n"
                                  "Please close at least one and retry." );
                    break;

                case ENFILE :
                    show_message( "Sorry, system limit for open files "
                                  "exceeded!\n Please try to close some "
                                  "files and retry." );
                break;

                case ENOSPC :
                    show_message( "Sorry, no space left on device for more "
                                  "file!\n    Please delete some files and "
                                  "retry." );
                    break;

                default :
                    show_message( "Sorry, can't open selected file for "
                                  "writing!\n       Please select a "
                                  "different file." );
            }

            name = T_free( name );
            continue;
        }
    } while ( rs690.dump_file == NULL );

    T_free( name );
}


/*-------------------------------------------------------------------*
 *-------------------------------------------------------------------*/

void
rs690_dump_channels( FILE * fp )
{
    Function_T *f;
    Channel_T *ch;
    int i, j, k;
    const char *plist[ ] = { "+X", "-X", "+Y", "-Y" };


    if ( fp == NULL )
        return;

    fprintf( fp, "===\n" );

    for ( i = 0; i < PULSER_CHANNEL_NUM_FUNC; i++ )
    {
        f = rs690.function + i;

        if ( ! f->is_needed && f->num_channels == 0 )
            continue;

        for ( j = 0; j < f->num_channels; j++ )
        {
            ch = f->channel[ j ];

            if ( ! ch->needs_update )
                continue;

            fprintf( fp, "%s:%s", f->name, rs690_num_2_channel( ch->self ) );

            for ( k = 0; k < ch->num_active_pulses; k++ )
            {
                if (    f->self == PULSER_CHANNEL_PULSE_SHAPE
                     && ch->pulse_params[ k ].pulse->sp != NULL )
                    fprintf( fp, " (%ld) %ld %ld",
                             ch->pulse_params[ k ].pulse->sp->num,
                             ch->pulse_params[ k ].pos,
                             ch->pulse_params[ k ].len );
                else if (    f->self == PULSER_CHANNEL_TWT
                          && ch->pulse_params[ k ].pulse->tp != NULL )
                    fprintf( fp, " (%ld) %ld %ld",
                             ch->pulse_params[ k ].pulse->tp->num,
                             ch->pulse_params[ k ].pos,
                             ch->pulse_params[ k ].len );
                else
                    fprintf( fp, " %ld %ld %ld",
                             ch->pulse_params[ k ].pulse->num,
                             ch->pulse_params[ k ].pos,
                             ch->pulse_params[ k ].len );

                if ( f->phase_setup != NULL )
                {
                    if ( ch->pulse_params[ k ].pulse->pc == NULL )
                        fprintf( fp, " +X" );
                    else
                        fprintf( fp, " %s",
                                 plist[ ch->pulse_params[ k ].pulse->pc->
                                                 sequence[ f->next_phase ] ] );
                }
            }

            fprintf( fp, "\n" );
        }
    }
}


/*-------------------------------------------------------------------*
 *-------------------------------------------------------------------*/

void
rs690_duty_check( void )
{
    Function_T *f;
    int i;
    int fns[ ] = { PULSER_CHANNEL_TWT, PULSER_CHANNEL_TWT_GATE };

    if ( ! rs690.is_repeat_time )
        return;


    for ( i = 0; i < 2; i++ )
    {
        f = rs690.function + fns[ i ];
        if (    f->is_used
             && f->num_channels > 0
             && rs690_calc_max_length( f ) >
                                        MAX_TWT_DUTY_CYCLE * rs690.repeat_time
             && f->max_duty_warning++ == 0 )
                print( SEVERE, "Duty cycle of TWT exceeded due to length of "
                       "%s pulses.\n", f->name );
    }
}


/*-------------------------------------------------------------------*
 *-------------------------------------------------------------------*/

Ticks
rs690_calc_max_length( Function_T * f )
{
    int i, j;
    Channel_T *ch;
    Ticks max_len = 0;


    if ( ! f->is_needed || f->num_channels == 0 )
        return 0;

    for ( j = 0; j < f->num_channels; j++ )
    {
        ch = f->channel[ j ];

        for ( i = 0; i < ch->num_active_pulses; i++ )
            max_len += ch->pulse_params[ i ].len;
    }

    return max_len;
}


/*-------------------------------------------------------------------*
 *-------------------------------------------------------------------*/

char *
rs690_num_2_channel( int num )
{
    static char name[ 4 ];


    sprintf( name, "%c%d", num / 16 + 'A', num % 16 );
    return name;
}


/*-----------------------------------------------*
 * This is a no-op function and only implemented
 * because some other pulsers also have it.
 *-----------------------------------------------*/

bool
rs690_set_max_seq_len( double seq_len  UNUSED_ARG )
{
    print( WARN, "Pulser doesn't allow setting a maximum pattern length.\n" );
    return OK;
}


/*-------------------------------------------------------------------*
 *-------------------------------------------------------------------*/

long
rs690_ch_to_num( long channel )
{
    if ( channel < CHANNEL_A0
         || ( channel > CHANNEL_D15 && NUM_HSM_CARDS == 1 )
         || ( channel > CHANNEL_H15 && NUM_HSM_CARDS == 2 ) )
    {
         print( FATAL, "Pulser has no channel named '%s'.\n",
                Channel_Names[ channel ] );
         THROW( EXCEPTION );
    }

    return channel - CHANNEL_A0;
}


/*
 * Local variables:
 * tags-file-name: "../TAGS"
 * tab-width: 4
 * indent-tabs-mode: nil
 * End:
 */
